
section {* Haskell Language *}

theory GHCCore
imports Main
begin

text {*
What type of markup do we use here? Is this LaTeX? Or Markdown? Plain text?
I'm assuming none of the above.

The HCore presented here is based on the "data Expr" in "CoreSyn.lhs" of GHC.
We will refer to the Haskell Core in GHC as HCore.
When refering to our version we just mention the relevant identifier.

Differences between expr and Expr in HCore:
  In expr, "some_identifier" is used for any kind of HCore-identifier:
    function symbols, variable names, constants, constructors etc all have the same type.
    The only distinction is between Literals and Variables.
  We cannot retrieve type information from expr, though it is present in HCore.
  In expr, we cannot know whether a constructor has all its fields filled in.
  Fortunately, none of the definitions depend on this.
  Cast, tick and type all do not influence execution of HCore.
    For this reason, they are represented by a simple string in expr.
  We will interpret (error "") to be a diverging computation,
    even though it is a finite computation for HCore. This is motivated by the observations that:
    - error "" is usually an undesired result in programs,
      so this interpretation may be acceptable in practice.
    - since error aborts an otherwise non-terminating sequence,
      catching errors would require us to build a special mechanism.
  In HCore, Let bindings are expected to be shared / call-by-need.
    In expr, this is not our interpretation. There is a note on this later.

An important difference between Haskell and HCore:
  Both in HCore and expr, case expressions are interpreted as strict.
  This is an intentional difference between HCore and Haskell, and we follow HCore in this.
  The following term would evaluate to [] in Haskell:

    case y@(error "") {_ -> []}
  
  The corresponding HCore/expr would simply be (Lit []), without mentioning y or error.
  The following term evaluates (error "") in HCore/expr:
    
    Case (error "") ''y'' [Alt DEFAULT Nil (Lit ''[]'')]
  
  In other words, HCore first evaluates the scrutinee.
  I.e. for the following HCore:

    case e1 litnm [alt1,alt2,...,altn]
  
  the corresponding Haskell program would be:
  
    seq e1 (case e1 litnm {altn;...;alt2;alt1})

  Note that the order of the alternatives also changes
  (For efficiency reason, Core was designed to always put the default first).

  The "forccee" example below illustrates why this is useful.

    Haskell
    
      map :: (a -> b) -> [a] -> [b]
      map _ []     = []
      map f (x:xs) = f x : map f xs
      
      forccee :: a -> b -> b
      forccee = seq
    
    HCore
    
      map :: forall a b. (a -> b) -> [a] -> [b]
      map =
        \ (@ a) (@ b) (f :: a -> b) (xs :: [a]) ->
          case xs of _ {
            []     -> GHC.Types.[] @ b;
            : y ys -> GHC.Types.: @ b (f y) (map @ a @ b f ys)
          }
      
      forccee = \ (x :: a) (y :: b) -> case x of _ { __DEFAULT -> y }
    
    expr: see GHCTest.thy

View core of "A" before optimizations: ghc -ddump-ds -fforce-recomp -O0 A.hs
View evaluation order: ghc -ddump-stg A.hs > A.stg

There is also a pretty-printer for ghc-core, called ghc-core, available in Cabal

*}

(* RT: why not?
datatype some_identifier = Idf string

or even

type_synonym some_identifier = string
*)
datatype some_identifier = Idf "char list"

(* RT: please avoid camelCase, so use alt_con, Lit_Alt, and Data_Alt *)
datatype altCon = LitAlt some_identifier | DataAlt some_identifier | DEFAULT

(* RT: why String.literal and not just string *)
(* RT: drop quotes aroung "bind", "expr", etc *)
(* RT: why not polymorphic type for identifiers *)
datatype expr
         = Var some_identifier (* free variable, or bound by a 'b *)
         | Lit some_identifier (* literal, elsewhere defined *)
         | App "expr" "expr" (* function application: first the function, second the argument *)
         | Lam some_identifier "expr" (* lambda abstraction / big-lambda (type) abstraction *)
         | Let "bind" "expr"
           (* Let is a lot like lambda abstraction together with function application,
              but it allows recursive calls within the binding *)
         | Case "expr" some_identifier "alt list"
           (* All pattern matching / evaluation happens via case.
              Note the extra binding 'b, which is the evaluation of the
              expression just before it.  *)
         | TickOrCast "expr" String.literal (* or coerce *)
         | Type String.literal
     and bind
         = NonRec some_identifier expr
         | Rec "(some_identifier \<times> expr) list"
         (* careful: NonRec b e \<noteq> Rec [(b,e)].
            Consider 'let x = x+1':
               this would be non-terminating in case of Rec,
               but it is terminating if x is in NonRec (it just 'updates' x with +1)
          *)
     and alt  = Alt altCon "some_identifier list" "expr"

datatype 'a executionResult
  = FreeVariable some_identifier (* Missing a binder for some_identifier *)
  | NotWellFormed (* An error occurred that shouldn't have: is this ghc-generated Core? *)
  | BindError (* One of the bindings overwrites another. Either the evaluateHead was called with non-nil bindings, or it has a bug *)
  | IsWHNF 'a (* The term is already in WHNF, nothing left to do *)
  | Step 'a (* Success: made a step *)

definition mplus_executionResult where
"mplus_executionResult a b = (case a of (Step v) \<Rightarrow> Step v | _ \<Rightarrow> b)"

lemma mplus_executionResult_simps[simp] :
shows "mplus_executionResult (Step a) b = Step a"
  and "mplus_executionResult (FreeVariable v) b = b"
  and "mplus_executionResult NotWellFormed b = b"
  and "mplus_executionResult BindError b = b"
  and "mplus_executionResult (IsWHNF w) b = b" by (simp_all add:mplus_executionResult_def)

lemma mplus_executionResult_associative :
"mplus_executionResult (mplus_executionResult a b) c
 = mplus_executionResult a (mplus_executionResult b c)"
by (cases a,simp_all)

fun unravelBinding :: "bind \<Rightarrow> (some_identifier \<times> expr) list" where
"unravelBinding (NonRec x v) = [(x,v)]" |
"unravelBinding (Rec xs) = xs"

fun findAssoc where
"findAssoc i Nil = FreeVariable i" |
"findAssoc i (Cons (x,e) xs) = (if x = i then Step e else findAssoc i xs)"

fun nmsize where "nmsize (Idf nm) = length nm"

(* RT: perhaps use "_ \<in> set _" *)
lemma listsum_larger_than_element':
  "Idf nm \<in> set lst \<Longrightarrow> listsum (map nmsize lst) \<ge> length nm"
  by (induct lst, auto)

lemma listsum_larger_than_element :
  "List.member lst (Idf nm) \<Longrightarrow> listsum (map nmsize lst) \<ge> length nm"
  by (induct lst, auto simp: member_rec)
(* RT:
  show "(List.member [] (Idf nm)) \<Longrightarrow> length nm \<le> listsum (map nmsize [])"
    by (simp add: member_rec(2))
  { fix a lst
    assume lst: "List.member lst (Idf nm) \<Longrightarrow> length nm \<le> listsum (map nmsize lst)"
    hence "List.member (a # lst) (Idf nm) \<Longrightarrow> length nm \<le> listsum (map nmsize (a # lst))"
    proof(cases "a = Idf nm",simp_all add: member_rec(1)) qed
  }
  thus "\<And>a lst. (List.member lst (Idf nm) \<Longrightarrow> length nm \<le> listsum (map nmsize lst)) \<Longrightarrow>
    List.member (a # lst) (Idf nm) \<Longrightarrow> length nm \<le> listsum (map nmsize (a # lst))" by auto
qed *)

lemma listsumSuc :
  assumes "List.member lst (Idf nm)"
  shows "listsum (map nmsize lst) - length nm < Suc (listsum (map nmsize lst)) - length nm"
proof -
  have "listsum (map nmsize lst) \<ge> length nm"
    using assms listsum_larger_than_element[of lst nm] by simp
  thus ?thesis by simp
qed

(*RT: reuse existing functionality to get fresh names *)
declare listsumSuc[termination_simp]
function findNewName :: "some_identifier \<Rightarrow> some_identifier list \<Rightarrow> some_identifier" where
"findNewName (Idf nm) lst
  = (case List.member lst (Idf nm) of
    True \<Rightarrow> findNewName (Idf (nm @ ''*'')) lst | 
    False \<Rightarrow> Idf nm)" by pat_completeness auto
termination by (relation "measure (%(nm,lst). (Suc (listsum (map nmsize lst)) - nmsize nm))"
               , simp_all add:listsumSuc )

fun newnameAdd where
"newnameAdd renamedAlready x = Cons (x, findNewName x (map snd renamedAlready)) renamedAlready"

fun findNewNames
  :: " (some_identifier \<times> some_identifier) list
     \<Rightarrow> some_identifier list \<Rightarrow> (some_identifier \<times> some_identifier) list"
  where
"findNewNames renamedAlready Nil = renamedAlready" |
"findNewNames renamedAlready (Cons x xs)
  = (findNewNames (newnameAdd renamedAlready x) xs)"

fun useRenaming where
"useRenaming renamings x = (case map_of renamings x of
                                      (Some v) \<Rightarrow> v | Nothing \<Rightarrow> x)"

lemma snd_of_pair_elim[simp]: "z2 \<in> Basic_BNFs.snds (a, b) \<longleftrightarrow> z2 = b"
by (simp add: snds.simps snds.cases) (* required for proving termination of makeFresh *)

lemma fst_of_pair_elim[simp]: "z2 \<in> Basic_BNFs.fsts (a, b) \<longleftrightarrow> z2 = a"
by (simp add: fsts.simps fsts.cases) (* not required, but symmetrical *)

fun makeFresh :: "(some_identifier \<times> some_identifier) list \<Rightarrow> expr \<Rightarrow> expr" where
"makeFresh renamings (Var x) = Var (useRenaming renamings x)" |
"makeFresh _ (Lit x) = Lit x" |
"makeFresh renamings (Let (NonRec x v) e)
  = (let newrms = newnameAdd renamings x
    ; newname = useRenaming newrms x in
    Let (NonRec newname (makeFresh renamings v)) (makeFresh newrms e))" |
"makeFresh renamings (Let (Rec xs) e)
  = (let newnames = findNewNames renamings (map fst xs)
    ; newxs = map (map_prod (useRenaming newnames) (makeFresh newnames)) xs
    in Let (Rec newxs) (makeFresh newnames e))" |
"makeFresh renamings (Lam x e)
  = (let newrms = newnameAdd renamings x
    ; newname = useRenaming newrms x in
    Lam newname (makeFresh newrms e))" |
"makeFresh renamings (Case v x alts)
  = (let newrms = newnameAdd renamings x
    ; newname = useRenaming newrms x 
    ; newalts = map (\<lambda> a. (case a of (Alt con nms e)
                    \<Rightarrow> let newrms2 = findNewNames newrms nms in
                    Alt con (map (useRenaming newrms2) nms) (makeFresh newrms2 e))) alts in
    Case (makeFresh renamings v) newname newalts
    )" |
"makeFresh renamings (TickOrCast e l)
  = TickOrCast (makeFresh renamings e) l" |
"makeFresh _ (Type l) = Type l" |
"makeFresh renamings (App a b)
  = App (makeFresh renamings a) (makeFresh renamings b)"

fun double :: "'a \<Rightarrow> 'a \<times> 'a" where "double a = (a,a)"

fun unravelBindings :: "some_identifier list \<Rightarrow> bind list \<Rightarrow> (some_identifier \<times> expr) list" where
"unravelBindings _ Nil = Nil" |
"unravelBindings nonFree (Cons (NonRec x v) xs)
  = Cons (x, makeFresh (map double (Cons x nonFree)) v) (unravelBindings (Cons x nonFree) xs)" |
"unravelBindings nonFree (Cons (Rec recLst) xs)
  = (map (apsnd (makeFresh (map double (map fst recLst @ nonFree)))) recLst)
  @ unravelBindings (map fst recLst @ nonFree) xs"

definition findBinding :: "some_identifier \<Rightarrow> bind list \<Rightarrow> expr executionResult" where
"findBinding i = findAssoc i \<circ> unravelBindings Nil"

fun overLappingAssoc :: "(some_identifier \<times> expr) list \<Rightarrow> (some_identifier \<times> expr) list \<Rightarrow> bool"
where
"overLappingAssoc Nil ls = False" |
"overLappingAssoc (Cons (i,_) xs) ls = (case findAssoc i ls of Step v \<Rightarrow> True
                                                | _ \<Rightarrow> overLappingAssoc xs ls)"

definition overlappingBinding where
"overlappingBinding b bs = overLappingAssoc (unravelBinding b) (unravelBindings Nil bs)"

(* get rid of some unnecessary let-expressions by using this instead of the constructor Let *)
fun rewrapLet :: "bind \<Rightarrow> expr \<Rightarrow> expr" where
"rewrapLet _ (Lit v) = Lit v" |
"rewrapLet _ (Type v) = Type v" |
"rewrapLet b (App x y)
  = (case rewrapLet b x of
    (Let _ _) \<Rightarrow> Let b (App x y) |
    otherX \<Rightarrow> (case rewrapLet b y of (Let _ _) \<Rightarrow> Let b (App x y) | otherY \<Rightarrow> App otherX otherY))" |
"rewrapLet b e = Let b e"

fun rewrapLets :: "expr \<Rightarrow> bind list \<Rightarrow> expr" where
"rewrapLets e Nil = e" |
"rewrapLets e (Cons x xs) = (rewrapLet x (rewrapLets e xs))"

fun stripLets :: "expr \<Rightarrow> (expr \<times> bind list)" where
"stripLets (Let b e) = (case stripLets e of (e2, bs) \<Rightarrow> (e2, Cons b bs))" |
"stripLets e = (e,[])"

(*
lemma stripThenRewrapLetsIsId[simp] : 
shows "rewrapLets (fst (stripLets e)) (snd (stripLets e)) = e"
by(induct e,simp_all add: case_prod_unfold)
*)

fun matchAndBindDataAlt ::
"expr \<Rightarrow> some_identifier \<Rightarrow> some_identifier list \<Rightarrow> (bind list) executionResult" where
"matchAndBindDataAlt (Lit v) i Nil = (if i = v then Step Nil else NotWellFormed)" |
"matchAndBindDataAlt (App x y) i (Cons y' x')
  = map_executionResult (Cons (NonRec y' y)) (matchAndBindDataAlt x i x')" |
"matchAndBindDataAlt (Let b e) i b2 = map_executionResult (Cons b) (matchAndBindDataAlt e i b2)" |
"matchAndBindDataAlt _ i _ = NotWellFormed" (* number of constructor-arguments don't match *)

(* in the case of a match, we keep the order:
     Const a b c = ((Const a) b) c
   is matched with: DataAlt Const [a',b',c'] = DataAlt Const (Cons a (Cons b (Cons c Nil)))
   note that we need to reverse the list for that!
*)
fun matchAlt :: "expr \<Rightarrow> alt \<Rightarrow> expr executionResult" where
"matchAlt v (Alt (LitAlt litid) lst e) (* match built-in literal, like 0 *)
   = (case stripLets v of
       (Lit v2,_) \<Rightarrow> (if v2 = litid then Step e else NotWellFormed) |
       _ \<Rightarrow> NotWellFormed)" |
"matchAlt v (Alt (DataAlt litid) lst e)
   = map_executionResult (rewrapLets e) (matchAndBindDataAlt v litid (rev lst))" |
"matchAlt v (Alt DEFAULT lst e) = Step e"

(* Note: in HCore, the DEFAULT comes first, so we match in opposite order *)
fun matchAlts :: "expr \<Rightarrow> alt list \<Rightarrow> expr executionResult" where
"matchAlts e [] = NotWellFormed" |
"matchAlts e (Cons x xs)
  = (case matchAlts e xs of Step v \<Rightarrow> Step v | NotWellFormed \<Rightarrow> matchAlt e x | other \<Rightarrow> other)"

lemma stripLetsDecr : "size (fst (stripLets x)) \<le> size x"
by(induct x,simp_all add: prod.case_eq_if)

fun nonRecFresh where
"nonRecFresh bindings someVar someVal forExpr
  = makeFresh (map (double \<circ> fst) (unravelBindings Nil bindings)) (Let (NonRec someVar someVal) forExpr)"

(* evaluateHead evaluates an expression in a certain scope.
If the expression is already in WHNF, the constructor IsWHNF is returned.
Other possible return types are:
Step p:
  managed to make some progress, p is the new program to be evaluated
FreeVariable x:
  could not reduce to WHNF because the value of a variable needs to be filled in
NotWellFormed:
  something occurred that shouldn't after type-checking,
  probably the original program is illegal.

The evaluateHead 'runTime' differs from the Haskell-runtime:
   variables are immediately replaced by their definition,
   instead of first evaluating the variable to WHNF.
   For termination, this does not matter.
   For complexity, it does matter: sharing of evaluated results typically reduces complexity. *)
fun evaluateHead :: "expr \<Rightarrow> bind list \<Rightarrow> expr executionResult" where
"evaluateHead (Var i) bs = findBinding i bs" (* unfold bound variable. See Note *) |
"evaluateHead (Lit e) _ = IsWHNF (Lit e)" (* literals cannot be unfolded: term already in WHNF *) |
"evaluateHead (App e1 e2) bs
  = (case (stripLets e1) of
       (Lam i e, rs) \<Rightarrow> Step (nonRecFresh bs i e2 (rewrapLets e rs)) |
       _ \<Rightarrow> (* the term 'f x' where f is a bound variable, is not WHNF: unfold f *)
            map_executionResult (\<lambda> v. App v e2) (evaluateHead e1 bs))" |
"evaluateHead (Lam i e) bs = IsWHNF (Lam i e)" (* missing argument, cannot reduce *) |
 (* Let does not evaluate anything, but only represents scope *)
"evaluateHead (Let b e) bs
  = (if overlappingBinding b bs then BindError
     else map_executionResult (rewrapLet b) (evaluateHead e (Cons b bs)))" |
"evaluateHead (Case e b alts) bs
  = (case (evaluateHead e bs) of
       IsWHNF _ \<Rightarrow> map_executionResult (nonRecFresh bs b e) (matchAlts e alts) |
       Step e2 \<Rightarrow> Step (Case e2 b alts) |
       other \<Rightarrow> other)" |
"evaluateHead (TickOrCast e _) bs = Step e" (* ignore *) |
"evaluateHead (Type e) _ = IsWHNF (Type e)" (* similar to a literal: cannot be unfolded *)


(* Note:
when a bound variable is replaced, it could still contain variables:

let x = 3 in (let y = x in (let x = 4 in y))  yields 3 in ghci
However, if we replace y by its definition, namely x, we get:
let x = 3 in (let y = x in (let x = 4 in x))  yields 4 in ghci
As 3\<noteq>4, we shouldn't replace y by its definition in this way.
However, this seems like the right order in which to evaluate.

How we fix this:
Do not allow overwrite by Let expressions. Core must be generated such that this does not occur.
   In core, let x = 3 in (let y = x in (let x = 4 in y)) would be translated as:
            let x = 3 in (let y = x in (let x2 = 4 in y))
   This requires throwing an error once we add an already bound variable on the stack (variable bs)
   Bindings overwriting each other can be checked syntactically (i.e. without evaluation).
   Consequently, we must rename variables to keep free variables free, i.e.:
   let f = (\ y . (let x = 4 in y)) in
      (let x = 3 in (f x))
   Rewrites to:
   let f = ... in
      (let x = 3 in (\ y . (let x* = 4 in y)))
   Unfortunately, this means that a rewrite step depends on non-local scope

*)


(* evaluate will fully evaluate the constructor, instead of just its head *)
fun evaluate :: "expr \<Rightarrow> bind list \<Rightarrow> expr executionResult" where
"evaluate (Let b e) bs
  = (if overlappingBinding b bs then BindError
     else map_executionResult (rewrapLet b) (evaluate e (Cons b bs)))" |
"evaluate (App a b) bs
  = (case evaluateHead (App a b) bs of
     IsWHNF _ \<Rightarrow> (case evaluate a bs of
                   Step v \<Rightarrow> Step (App v b)
                 | IsWHNF _ \<Rightarrow> map_executionResult (App a) (evaluate b bs))
     | v \<Rightarrow> v
    )" |
"evaluate other bs
  = evaluateHead other bs"

(* The function 'evaluate' is tested in GHCTest *)

end
